(in-package :feuer)


;;
;; * a mesh consists of patches. one patch has one texture and one shader
;; * each vertex can only be a member of many polygons and many patches
;; * each vertex has texture coords stored in a plist, with the patch as key
;; * only triangle-patches can have normals associated with their polygons
;; * the reinit method is to be called when a new opengl context is created.
;;
;; todo: mesh muss allgemeinem composite zum opfer fallen, in dem dann patches drin sind.
;;


(defmacro with-checked-error (&body body)
  (let ((name (if (stringp (first body))
                  (prog1
                      (first body)
                    (setf body (rest body)))
                  "")))
    (with-gensyms (error-occured? error)
      `(let ((,error-occured? nil ))
         ,@(mapcar (lambda (stuff)
                     (if (and (listp stuff) (> (length stuff) 1) (member (first stuff) '(let let* when unless)))
                         `(,(first stuff) ,(second stuff)
                            (with-checked-error ,@(rest (rest stuff))))
                         `(prog1
                              ,stuff
                            (let ((,error (gl:get-error)))
                              (unless (or ,error-occured? (eq ,error :zero))
                                (setf ,error-occured? t)
                                (warn (format nil "~%###~a: ~a during ~a~%~%" ,name ,error ',stuff))
                                (force-output))))))
                   body)))))

(defclass* vertex ()
  ((pos)
   (polygons ())
   (normal nil)
   (array-index nil)
   (color nil)
   (tex-coords ())))

(defclass* polygon ()
  ((vertices ())
   (patch nil)))

(defclass* quad (polygon)
  ())

(defclass* triangle (polygon)
  ((normal nil)))


(defclass* patch ()
  ((vertices ())
   (indices ())
   (polygons ())
   (texture ())
   (shader ())
   (vertices-array nil)
   (indices-array nil)))

(defclass* mesh ()
  ((patches nil)
   (display-list nil)))

(defclass* quad-patch (patch)
  ())

(defclass* triangle-patch (patch)
  ())

(defun make-point (x y &optional z)
  (if z
      (vector3 (coerce x 'double-float)
               (coerce y 'double-float)
               (coerce z 'double-float))
      (vector2 (coerce x 'double-float)
               (coerce y 'double-float))))



(defun make-vertex (x y z &key (normal nil) (tex-coords nil))
  (make-instance 'vertex :pos (vector3 (coerce x 'double-float)
                                        (coerce y 'double-float)
                                        (coerce z 'double-float))
                         :normal normal
                         :tex-coords tex-coords))

(defmethod print-object ((vertex vertex) stream)
  (format stream "#<V~8,4f ~8,4f ~8,4f>"
          (x (pos-of vertex))
          (y (pos-of vertex))
          (z (pos-of vertex))))


(defmethod make-polygon ((vertices cons))
  (make-instance (case (length vertices)
                   (3 'triangle)
                   (4 'quad)
                   (t (error "polygon must be of length 3 or 4")))
                 :vertices vertices))


(defmethod v1 ((polygon polygon))
  (first (vertices-of polygon)))

(defmethod v2 ((polygon polygon))
  (second (vertices-of polygon)))

(defmethod v3 ((polygon polygon))
  (third (vertices-of polygon)))

(defmethod v4 ((polygon quad))
  (fourth (vertices-of polygon)))

(defmethod p1 ((polygon polygon))
  (pos-of (v1 polygon)))

(defmethod p2 ((polygon polygon))
  (pos-of (v2 polygon)))

(defmethod p3 ((polygon polygon))
  (pos-of (v3 polygon)))

(defmethod p4 ((polygon quad))
  (pos-of (v4 polygon)))

(defmethod tex-coord-of ((vertex vertex) (patch patch))
  (let ((tex-coord (getf (tex-coords-of vertex) patch)))
    (assert tex-coord)
    tex-coord))

(defmethod (setf tex-coord-of) (new-coord (vertex vertex) (patch patch))
  (setf (getf (tex-coords-of vertex) patch) new-coord))

(defmethod texture-of ((mesh mesh))
  (assert (= (length (patches-of mesh)) 1))
  (texture-of (first (patches-of mesh))))

(defmethod (setf texture-of) (new-texture (mesh mesh))
  (dolist (patch (patches-of mesh))
    (setf (texture-of patch) new-texture)))

(defmethod initialize-instance :after ((triangle triangle) &key normal)
  (unless normal
    (setf (normal-of triangle)
          (cross-product (normalize (m- (p2 triangle) (p1 triangle)))
                         (normalize (m- (p3 triangle) (p1 triangle)))))))

(defmethod initialize-instance :after ((poly polygon) &key vertices)
  (dolist (vertex vertices)
    (push poly (polygons-of vertex))))

(defmethod initialize-instance :after ((patch patch) &key polygons)
  (dolist (poly polygons)
    (setf (patch-of poly) patch)))


(defgeneric triangulate (thing &key))

(defmethod triangulate ((quad quad) &key (mode :fast) (new-triangle-class 'triangle))
  (if (eq mode :fast)
      (list (make-instance new-triangle-class
                           :vertices (list (v1 quad) (v2 quad) (v4 quad))
                           :patch (patch-of quad))
            (make-instance new-triangle-class
                           :vertices (list (v2 quad) (v3 quad) (v4 quad))
                           :patch (patch-of quad)))
      (error "not implemented")))

(defmethod triangulate ((patch quad-patch) &key (mode :fast)
                                                     (new-class 'triangle-patch)
                                                     (new-triangle-class 'triangle))
  (setf (polygons-of patch)
        (loop for polygon in (polygons-of patch)
              append (triangulate polygon :mode mode :new-triangle-class new-triangle-class)))
  (change-class patch new-class))

(defmethod triangulate ((mesh mesh) &key (mode :fast) (new-class 'triangle-patch) (new-triangle-class 'triangle))
  (dolist (patch (patches-of mesh))
    (triangulate patch :mode mode :new-class new-class :new-triangle-class new-triangle-class)))


(defun normal (vertex)
  (gl:normal (x (normal-of vertex))
             (y (normal-of vertex))
             (z (normal-of vertex))))

(defun vertex (vertex)
  (gl:vertex (x (pos-of vertex))
             (y (pos-of vertex))
             (z (pos-of vertex))))

(defun tex-coord (vertex patch)
  #+nil(assert (and (<= 0 (x (tex-coord-of vertex patch)))
               (<= 0 (y (tex-coord-of vertex patch)))
               (>= 1 (x (tex-coord-of vertex patch)))
               (>= 1 (y (tex-coord-of vertex patch)))))
  (gl:tex-coord (x (tex-coord-of vertex patch))
                (y (tex-coord-of vertex patch))))

(defgeneric draw (thing &key))

(defmethod draw ((triangle triangle)&key)
  (gl:with-primitives :triangles
    (dolist (vertex (vertices-of triangle))
      (when (normal-of vertex) (normal vertex))
      (when (color-of vertex) (color (color-of vertex)))
      (when (tex-coord-of vertex (patch-of triangle)) (tex-coord vertex (patch-of triangle)))
      (vertex vertex))))

(defmethod draw ((quad quad)&key)
  (declare (optimize (debug 3)))
  (gl:with-primitives :quads
    (dolist (vertex (vertices-of quad))
      (when (normal-of vertex) (normal vertex))
      (when (color-of vertex) (color (color-of vertex)))
      (when (tex-coord-of vertex (patch-of quad)) (tex-coord vertex (patch-of quad)))
      (vertex vertex))))

(defmethod draw ((patch patch) &key (use-texture t))
 ; (gl:color 1 1 1 1)
  (when (and use-texture (texture-of patch))
    (bind-texture (texture-of patch)))
  (mapcan #'draw (polygons-of patch)))

(defmethod draw ((mesh mesh) &key (make-display-list t) (use-texture t))
  (gl:bind-texture :texture-2d 0)
  (if  (or nil (display-list-of mesh))
      (gl:call-list (display-list-of mesh))
      (if (or nil make-display-list)
          (make-display-list mesh)
          (mapcan (lambda (patch)
                    (draw patch :use-texture use-texture)) (patches-of mesh)))))

(defgeneric guess-normal (vertex &optional mode))
(defmethod guess-normal ((vertex vertex) &optional (mode :mean))
  (setf (normal-of vertex)
        (case mode
          (:mean (normalize (reduce #'m+
                                    (loop for polygon in (polygons-of vertex)
                                          collect (normal-of polygon)))))
          (otherwise (error "not implemented")))))


(defmethod guess-normals ((patch triangle-patch))
  (dolist (vertex (vertices-of patch))
    (guess-normal vertex)))

(defmethod guess-normals ((mesh mesh))
  (mapcan #'guess-normals (patches-of mesh)))


(defmethod update-topology ((mesh mesh))
  "reconstruct the patch from polygons only"
  (dolist (patch (patches-of mesh))
    (dolist (polygon (polygons-of patch))
      (setf (patch-of polygon) patch)
      (dolist (vertex (vertices-of polygon))
        (setf (polygons-of vertex) ()))))
  (dolist (patch (patches-of mesh))
    (setf (vertices-of patch) ;sammle vertices unique in liste. ausserdem jedem vertex sagen in welchen polygons es ist
          (remove-duplicates (loop for polygon in (polygons-of patch)
                                   do (mapcar (lambda (vertex)
                                                (push polygon (polygons-of vertex)))
                                              (vertices-of polygon))
                                   append (vertices-of polygon))))))


(defmethod vertices-of ((mesh mesh))
  (loop for patch in (patches-of mesh)
        append (vertices-of patch)))

(defun make-display-list (mesh)
  (when (display-list-of mesh)
    (gl:delete-lists (display-list-of mesh) 1))
  (setf (display-list-of mesh) nil)
  (let ((new-list (gl:gen-lists 1)))
    (gl:with-new-list (new-list :compile)
      (gl:with-pushed-matrix
        (draw mesh :make-display-list nil)))
    (setf (display-list-of mesh) new-list)))

(defmethod reinit ((patch patch))
  (when (texture-of patch) (reinit (texture-of patch))))

(defmethod reinit ((mesh mesh))
  (setf (display-list-of mesh) nil)
  (mapcan #'reinit (patches-of mesh)))

;(defun update-vertex-array (array index pos)
;;  (format t "updating vertex ~a at index ~a..." pos index)
;  (loop for coord in '(x y z)
;        for i from 0 to 3
;        do (setf (gl:glaref array index coord)
;                 (coerce (nth i (data pos)) 'single-float))))
  



#|(defmethod update-topology ((patch triangle-patch))
  "reconstruct the patch from its polygons only"

  (gl:enable-client-state :vertex-array)  ;todo: move elsewhere
  (gl:enable-client-state :color-array)   ;todo: move elsewhere
  (debug-out "fange an")
  (debug-out (gl:get-error))
 
  (dolist (polygon (polygons-of patch))
    (dolist (vertex (vertices-of polygon))
      (setf (polygons-of vertex) ())))
  (when (vertices-array-of patch)
    (gl:delete-buffers (list (vertices-array-of patch))))
  (when (indices-array-of patch)
    (gl:delete-buffers (list (indices-array-of patch))))
  
  ;;erstmal alle vertices unique machen
  ;;dann fuer jedes dreieck drei indizes da rein machen
  ;;eigentlich gleich beim unique-liste bauen indizes merken? nein, spaeter.

  (debug-out "foo")
  (debug-out (gl:get-error))
  (setf (vertices-of patch) ;sammle vertices unique in liste. ausserdem jedem vertex sagen in welchen polygons es ist
        (remove-duplicates (loop for polygon in (polygons-of patch)
                                 do (mapcar (lambda (vertex)
                                              (push polygon (polygons-of vertex)))
                                            (vertices-of polygon))
                                 append (vertices-of polygon)))
        
        (indices-of patch) ;indices sammeln. todo: das muss nicht in n^2 sein
        (loop for polygon in (polygons-of patch) 
              append  (mapcar (lambda (vertex)
                                (position vertex (vertices-of patch)))
                              (vertices-of polygon)))
        
        (vertices-array-of patch) ;generate names for vertex and index buffers
        (first (gl:gen-buffers 1))
        
        (indices-array-of patch)
        (first (gl:gen-buffers 1)))
  (debug-out "bar")
  (debug-out (gl:get-error))
  ;;setup the VBOs and fill them with initial data
  (let ((vertices (gl:alloc-gl-array 'position-only (length (vertices-of patch))))
        (indices (gl:alloc-gl-array :unsigned-int  (length (indices-of patch)))))
    ;;collect data into gl array
    (loop for i from 0 below (length (vertices-of patch))
          for vertex in (vertices-of patch)
          do (progn
               (loop for coord in '(x y z) ;das muss besser gehen, alles auf einmal
                     do (setf (gl:glaref vertices i coord) 
                              (coerce (funcall coord (pos-of vertex)) 'single-float)))
                                        ;(loop for color in '(r g b a) ;das muss besser gehen, alles auf einmal
                                        ;      do (setf (gl:glaref vertices i color) 
                                        ;              (random 255)))
               (setf (array-index-of vertex) i)))
    (debug-out "blafasel")
    (debug-out (gl:get-error))
    (loop for i from 0 below (length (indices-of patch))
          for index in (indices-of patch)
          do (setf (gl:glaref indices i) index))
    ;;bind buffers, initialize with contents of arrays
    (gl:bind-buffer :array-buffer (vertices-array-of patch))
    (gl:buffer-data :array-buffer :static-draw vertices)
  
    (gl:bind-buffer :element-array-buffer (indices-array-of patch))
    (gl:buffer-data :element-array-buffer :static-draw indices)
    ;; unbind buffers, free arrays. buffers should now contain the data.
  
    (gl:bind-buffer :array-buffer 0)
    (gl:bind-buffer :element-array-buffer 0)

    (gl:free-gl-array vertices)
    (gl:free-gl-array indices)
    (debug-out "und fertig")
    (debug-out (gl:get-error))))|#



(defun init ()
  (with-checked-error "init"
    (gl:enable-client-state :vertex-array) ;todo: move elsewhere
    (gl:enable-client-state :color-array)  ;todo: move elsewhere
    (let ((v-name (first (gl:gen-buffers 1)))
          (i-name (first (gl:gen-buffers 1))))
      (gl:bind-buffer :array-buffer v-name)
      (gl:bind-buffer :element-array-buffer i-name)
      (force-output)
      (let ((vertices (gl:alloc-gl-array 'position-only 3))
            (indices (gl:alloc-gl-array :unsigned-int  3)))
        (dotimes (i 3)
          (setf (gl:glaref vertices i 'x) (+ i 0.0f0)
                (gl:glaref vertices i 'y) (+ i 1.0f0)
                (gl:glaref vertices i 'z) (+ i 2.0f0)
                (gl:glaref indices  i  ) i))
                                        ;      (debug-out (gl:gl-array-byte-size vertices))
                                        ;      (debug-out (gl:gl-array-byte-size indices))
        (gl:buffer-data :array-buffer :static-draw vertices :size (gl:gl-array-byte-size vertices))
        (gl:buffer-data :element-array-buffer :static-draw indices :size (gl:gl-array-byte-size indices))
                                        ;  (gl:unmap-buffer :array-buffer)
                                        ;  (gl:unmap-buffer :element-array-buffer)
        (gl:free-gl-array vertices)
        (gl:free-gl-array indices)))))

(defun foo ()
  (with-checked-error "foo"
    (gl:color 1 1 1 1)
    (gl:enable-client-state :vertex-array)
    (gl:bind-buffer :array-buffer 1)
    (%gl:vertex-pointer 3 :float 0 (cffi-sys:make-pointer 0))
    (gl:enable-client-state :index-array)
    (gl:bind-buffer :element-array-buffer 1)
    (%gl:draw-elements :triangles 3 :unsigned-int (cffi-sys:make-pointer 0))))
 
  ; (gl:with-gl-mapped-buffer (vertices :array-buffer :write-only 'position-color (length (vertices-of patch)))
 ;   (declare (ignore vertices))
 ;   (debug-out (gl:get-error))
 ;   (gl:with-gl-mapped-buffer (indices :element-array-buffer :write-only :unsigned-int (* 0 (length (indices-of patch))))
 ;     (debug-out (gl:get-error))
 ;     (gl:draw-elements :triangles indices)))
  
                          
  




(defmethod draw-doesnotwork ((patch patch))
  (with-checked-error
    (debug-out "drawing patch")
    (debug-out "-----------")
  
 
 
  
    (gl:bind-buffer :array-buffer 1)
    (gl:bind-buffer :element-array-buffer 1)
    (gl:enable-client-state :vertex-array)
    (%gl:vertex-pointer 3 :float 0 (cffi-sys:make-pointer 0))
  
  
 ;; (gl:enable-client-state :index-array)

  
    (%gl:draw-elements :triangles 3 :unsigned-byte (cffi-sys:make-pointer 0))
 
  ; (gl:with-gl-mapped-buffer (vertices :array-buffer :write-only 'position-color (length (vertices-of patch)))
 ;   (declare (ignore vertices))
 ;   (debug-out (gl:get-error))
 ;   (gl:with-gl-mapped-buffer (indices :element-array-buffer :write-only :unsigned-int (* 0 (length (indices-of patch))))
 ;     (debug-out (gl:get-error))
 ;     (gl:draw-elements :triangles indices)))
  
                          
    ))


;(gl:define-gl-array-format position-color
;  (gl:vertex :type :float :components (x y z))
;  (gl:color :type :unsigned-char :components (r g b a)))


(gl:define-gl-array-format position-only
  (gl:vertex :type :float :components (x y z)))










